from django.contrib.auth import get_user_model

from .models import Ticket

User = get_user_model()


def get_pending_tickets(request):
    return {
        "pending_tickets": Ticket.objects.filter(status="pending").count(),
    }



import difflib
from django.utils.translation import gettext as _
from rest_framework import serializers, exceptions
from .models import Ticket, TicketMessage, TicketOptions


class CreateTicketSerializer(serializers.ModelSerializer):
    message = serializers.CharField(required=True)

    class Meta:
        model = Ticket
        fields = ("user", "status", "title", "section", "priority", "seen_by_admin", "seen_by_user", "message")

    def validate(self, attrs):
        valid = super().validate(attrs)
        valid["seen_by_user"] = True
        if not attrs.get("priority"):
            valid["priority"] = TicketOptions.LOW
        if not attrs.get("section"):
            valid["section"] = TicketOptions.SUPPORT
        return valid


class CreateTicketAPIViewSerializer(serializers.ModelSerializer):
    message = serializers.CharField(required=True)

    class Meta:
        model = Ticket
        fields = ("title", "section", "priority", "message")


class AddMessageSerializer(serializers.ModelSerializer):
    class Meta:
        model = TicketMessage
        fields = ("user", "ticket", "message")

    def validate(self, attrs):
        valid = super().validate(attrs)
        ticket = attrs.get("ticket")
        user = attrs.get("user")
        message = attrs.get("message", None)

        if ticket.user.id != user.id:
            raise exceptions.PermissionDenied()

        if ticket.status == TicketOptions.CLOSED:
            raise exceptions.ValidationError(_("Ticket has been closed."))

        if any([difflib.SequenceMatcher(None, message, x.message).ratio() > 0.85 for x in
                TicketMessage.objects.filter(ticket=ticket)]):
            raise exceptions.ValidationError(
                {"message": _("Message with similarity of 85% has already been sent.")})

        return valid


class TicketSerializer(serializers.ModelSerializer):
    class Meta: 
        model = Ticket
        fields = '__all__'


class TicketMessageSerializer(serializers.ModelSerializer):
    class Meta:
        model = TicketMessage
        fields = "__all__"


class TicketDetailSerializer(serializers.ModelSerializer):
    ticket_messages = serializers.SerializerMethodField(read_only=True)
    class Meta: 
        model = Ticket
        fields = '__all__'
    
    def get_ticket_messages(self, obj):
        messages = obj.ticketmessage_set.filter(soft_delete=False).order_by('id')
        return TicketMessageSerializer(messages, many=True, read_only=True).data
    
    def to_representation(self, instance):
        rep = super().to_representation(instance)
        for message in rep['ticket_messages']:
            if message['user'] == rep['user']:
                message['is_admin'] = False
            else:
                message['is_admin'] = True
        return rep





from django.dispatch import receiver
from django.db.models.signals import post_save 
from .models import TicketMessage , TicketOptions
from datetime import datetime
@receiver(post_save, sender=TicketMessage)
def after_create_ticket_message(sender, instance, created, **kwargs):
    if created and instance.ticket:
        if instance.user and (instance.user.is_superuser or instance.user.is_staff):
            instance.ticket.status = TicketOptions.ANSWERED
            instance.ticket.seen_by_user = False
            instance.ticket.seen_by_admin =  True
            instance.viewed=datetime.now()
            instance.viewer=instance.user
            instance.save()
        else:
            instance.ticket.status = TicketOptions.PENDING
            instance.ticket.seen_by_user = True
            instance.ticket.ticketmessage_set.filter(viewed__isnull=True).exclude(user=instance.ticket.user.id).update(viewed=datetime.now(),viewer=instance.ticket.user.id)
    
        instance.ticket.save()



from django.contrib import admin
from .models import Ticket, TicketMessage
# Register your models here.
from django.contrib.auth import get_user_model
from django.contrib import messages
from datetime import datetime
from django import forms
User = get_user_model()

class TicketMessageForm(forms.ModelForm):
    class Meta:
        model = TicketMessage
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk:
            self.fields["soft_delete"].disabled = False
            self.fields["message"].disabled = True
        else : self.fields["soft_delete"].disabled = True

class UnitInline(admin.TabularInline):
    model = TicketMessage
    form = TicketMessageForm
    extra = 1
    can_delete = False
    readonly_fields =  ("user","get_is_admin","created","viewer","viewed")

    def has_add_permission(self, request, obj=None): return True

    def get_is_admin(self,obj):
        if obj.user.is_superuser or obj.user.is_staff : return True
        return False
    get_is_admin.boolean = True
    get_is_admin.short_description = "is admin"


@admin.register(Ticket)
class TicketAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'title', 'status', 'priority', 'section','created','seen_by_admin','seen_by_user')
    list_display_links = list_display
    inlines = [UnitInline]
    
    def get_readonly_fields(self, request, obj=None):
        if obj:  # Only apply readonly for existing objects (change view)
            return ["user","title","section","priority","seen_by_user","seen_by_admin","created","updated","service","service_type"]
        return []
    
    def get_exclude(self, request, obj):
        if not obj: return ["seen_by_user","seen_by_admin","deleted_at"]
        return []
    add_fieldsets = (
        (None, {
            'fields': ('user', 'status',"title","section","priority"),
        }),
    )





    def render_change_form(self, request, context, *args, **kwargs):
        
        context.update({'save_text': True})
        return super().render_change_form(request, context, *args, **kwargs)
    

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)

        for instance in instances:
            if instance.pk: 
                original_instance = instance.__class__.objects.get(pk=instance.pk) 
                if instance.soft_delete != original_instance.soft_delete:
                    original_instance.soft_delete = instance.soft_delete
                    original_instance.save()
            else:
                instance.user = request.user
                instance.save()
        formset.save_m2m()

    def change_view(self, request, object_id, form_url='', extra_context=None):
        obj = self.get_object(request, object_id)
        if obj and not obj.seen_by_admin:
            obj.seen_by_admin = True
            obj.save()
            messages.add_message(request, messages.INFO, 'Field seen_by_admin has been updated to True.')
        obj.ticketmessage_set.filter(viewed__isnull=True , user=obj.user).update(viewed=datetime.now(),viewer=request.user)
        
        return super().change_view(request, object_id, form_url, extra_context)


from setuptools import setup

try:
    import pypandoc

    long_description = pypandoc.convert_file("README.md", "rst")
except (IOError, ImportError):
    long_description = open("README.md").read()

setup(
    url="https://github.com/HosseinSayyedMousavi/django-ticket",
    name="django-ticket",
    version="3.1.3",
    description="A ticket application for Django projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Hossein Sayyedmousavi",
    maintainer_email="Hossein.Sayyedmousavi@gmail.com",
    author="Abbas Ebadian",
    author_email="AbbasEbadiann@gmail.com",
    keywords=[
        "django",
        "database",
        "ticket",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Environment :: Console",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Database",
        "Topic :: System :: Archiving",
        "Topic :: System :: Archiving :: Compression",
    ],
    packages=["ticket"],
    install_requires=[
        "django",
        "djangorestframework",
        "drf-spectacular"
    ]
)



import datetime
from datetime import timezone
from enum import Enum

from django.contrib.auth import get_user_model
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext as _

User = get_user_model()


class TicketOptions(Enum):
    # status
    PENDING = "pending"
    PENDING_LABEL = _("Pending")
    ANSWERED = "answered"
    ANSWERED_LABEL = _("Answered")
    CLOSED = "closed"
    CLOSED_LABEL = _("Closed")

    # section
    MANAGEMENT = "management"
    MANAGEMENT_LABEL = _("Management")
    FINANCES = "finances"
    FINANCES_LABEL = _("Finances")
    SUPPORT = "support"
    SUPPORT_LABEL = _("Support")

    # priority
    LOW = "low"
    LOW_LABEL = _("Low")
    MEDIUM = "medium"
    MEDIUM_LABEL = _("Medium")
    HIGH = "high"
    HIGH_LABEL = _("High")


class Ticket(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    title = models.CharField(max_length=255)
    status = models.CharField(
        max_length=255,
        choices=[
            (TicketOptions.PENDING.value, TicketOptions.PENDING_LABEL.value),
            (TicketOptions.ANSWERED.value, TicketOptions.ANSWERED_LABEL.value),
            (TicketOptions.CLOSED.value, TicketOptions.CLOSED_LABEL.value),
        ],
        default=TicketOptions.PENDING,
    )
    section = models.CharField(
        max_length=128,
        choices=[
            (TicketOptions.MANAGEMENT.value, TicketOptions.MANAGEMENT_LABEL.value),
            (TicketOptions.FINANCES.value, TicketOptions.FINANCES_LABEL.value),
            (TicketOptions.SUPPORT.value, TicketOptions.SUPPORT_LABEL.value),
        ],
        default=TicketOptions.SUPPORT.value,
    )

    priority = models.CharField(
        max_length=128,
        choices=[
            (TicketOptions.LOW.value, TicketOptions.LOW_LABEL.value),
            (TicketOptions.MEDIUM.value, TicketOptions.MEDIUM_LABEL.value),
            (TicketOptions.HIGH_LABEL.value, TicketOptions.HIGH_LABEL.value),
        ],
        default=TicketOptions.LOW.value,
    )
    seen_by_user = models.BooleanField(default=False, verbose_name=_("Seen by user"))
    seen_by_admin = models.BooleanField(default=False, verbose_name=_("Seen by admin"))

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    soft_delete = models.BooleanField(default = False)
    def __str__(self):
        return f"{self.id}:{self.title}"

    class Meta:
        verbose_name = _("Ticket")
        verbose_name_plural = _("Tickets")


class TicketMessage(models.Model):
    ticket = models.ForeignKey(Ticket, on_delete=models.CASCADE, null=False, blank=False)
    user = models.ForeignKey(User, on_delete=models.CASCADE, null=False, blank=False)
    message = models.TextField(null=False, blank=False)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    viewed = models.DateTimeField(null=True)
    viewer = models.ForeignKey(User, on_delete=models.CASCADE, null=True , related_name='admin_viewer')
    soft_delete = models.BooleanField(default=False)
    def __str__(self):
        return ""

    class Meta:
        verbose_name = _("Ticket Message")
        verbose_name_plural = _("Ticket Messages")



from rest_framework.response import Response
from .models import Ticket, TicketMessage , TicketOptions
from django.contrib.auth import get_user_model
from rest_framework import generics ,permissions
from rest_framework.pagination import PageNumberPagination
from .serializers import   (TicketSerializer , CreateTicketSerializer , TicketDetailSerializer , AddMessageSerializer)
from rest_framework import status
from drf_spectacular.utils import extend_schema, OpenApiExample
from datetime import datetime
from django.utils.translation import gettext as _
User = get_user_model()


class ListPagination(PageNumberPagination):
    page_size = 10 
    page_size_query_param = 'page_size'
    max_page_size = 100


class TicketAPIView(generics.ListAPIView):
    permission_classes =[permissions.IsAuthenticated]
    serializer_class = TicketSerializer
    pagination_class = ListPagination

    def get_queryset(self):
        queryset = Ticket.objects.filter(user = self.request.user.id,soft_delete=False).order_by('-id')
        return queryset
    
    @extend_schema(
        description=_("Get list of user's tickets"),
    )
    def get(self,*args,**kwargs):
        return super().get(*args,**kwargs)
        
    @extend_schema(
        responses={status.HTTP_201_CREATED: ''},
        description=_('Create a ticket from user'),
        examples=[
            OpenApiExample(
                name=_('body example'),
                value={
                            'title': 'string Maxlength=50 , Minlength=10',
                            'section': "'management' or 'finances' or 'technical'",
                            'priority': "'low' or 'medium' or 'high'",
                            'message': 'Maxlength=1000',
                            },
                request_only=True,
            )
        ]
    )
    def post(self,request):
        request.data['user']=request.user.id
        serializer = CreateTicketSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        message = {'message':serializer.validated_data['message'],
                   'user':serializer.validated_data['user']}
        serializer.validated_data.pop('message')
        ticket = Ticket.objects.create(**serializer.validated_data)
        ticket.ticketmessage_set.create(**message)
        return Response({'detail': _('Ticket successfully submitted.')},status=status.HTTP_201_CREATED)


class TicketDetailAPIView(generics.RetrieveAPIView):
    permission_classes =[permissions.IsAuthenticated]
    serializer_class = TicketDetailSerializer
    lookup_field = 'id' 

    def get_queryset(self):
        queryset = Ticket.objects.filter(user = self.request.user.id,soft_delete=False).order_by('-id')
        return queryset

    @extend_schema(
        description=_("Get a specific ticket detail with it's id and seen it"),
    )
    def get(self,*args,**kwargs):
        ticket = self.get_object()
        ticket.seen_by_user = True
        ticket.ticketmessage_set.filter(viewed__isnull=True,user=ticket.user.id).update(viewed=datetime.now(),viewer=ticket.user.id)
        ticket.save()
        return super().get(*args,**kwargs)
    
    @extend_schema(
        responses={status.HTTP_201_CREATED:_('Your message has been successfully submitted.')},
        description='Create a message on a ticket of user',
        examples=[
            OpenApiExample(
                name='body example',
                value={
                            'message': 'Maxlength=1000'
                            },
                request_only=True,
            )
        ]
    )
    def post(self,request,id):
        request.data['user']=request.user.id
        request.data['ticket']=id
        serializer = AddMessageSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        TicketMessage.objects.create(**serializer.validated_data)

        return Response({'error': False, 'detail': _('Your message has been successfully submitted.')})

    @extend_schema(
        responses={status.HTTP_200_OK: _('Ticket closed.')},
        description=_('close a ticket'),
        request=None
    )
    def patch(self,request,id):
        ticket = self.get_object()
        ticket.status = TicketOptions.CLOSED
        ticket.save()
        return Response({'detail': _('Ticket closed.')})
